import weka.attributeSelection.*;
import weka.core.*;
import weka.core.converters.ConverterUtils.*;
import weka.classifiers.*;
import weka.classifiers.meta.*;
import weka.classifiers.trees.*;
import weka.filters.*;

import java.util.*;
import weka.core.converters.ConverterUtils.DataSource;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;

public class PCA {

    // Read a given dataset file (file must end with ".arff" and it should not be specified in the name parameter)
    protected Instances readData(String inputDir, String name) throws Exception {
        DataSource source = new DataSource(inputDir + "/" + name + ".arff");
        Instances data = source.getDataSet();
        // Make the last attribute be the class 
        if (data.classIndex() == -1) {
            data.setClassIndex(data.numAttributes() - 1);
        }
        return data;
    }

    // Executes Principal Component Analysis using a given dataset
    protected PrincipalComponents applyPCA(Instances data, double varianceCovered) throws Exception {
        AttributeSelectedClassifier classifier = new AttributeSelectedClassifier();
        
        // Create a Weka AS Evaluation algorithm: Principal Component Analysis
        PrincipalComponents pca = new PrincipalComponents();
        pca.setVarianceCovered(varianceCovered);
        
        // Create a Weka's AS Search algorithm: Ranker
        ASSearch search = new Ranker();
        
        AttributeSelection attrSel = new AttributeSelection();
        attrSel.setEvaluator(pca);
        attrSel.setSearch(search);

        // Perform the attribute selection
        attrSel.SelectAttributes(data);
        
        return pca;
    }
    
    // Saves Principal Component Analysis results into a file (it will add ".txt" extention automatically)
    protected void saveResults(String outputDir, String name, PrincipalComponents pca) throws Exception {
        BufferedWriter writer = new BufferedWriter(new FileWriter(outputDir + "/" + name + ".txt"));
        writer.write(pca.toString());
        writer.newLine();
        writer.flush();
        writer.close();
    }
    
    // Prints principal components
    protected void printPCA(String name, PrincipalComponents pca) throws Exception {
        System.out.println("------------------------------------------------------------");
        System.out.println(name);
        System.out.println("------------------------------------------------------------");
        
        // Access to principal components
        Instances headerAttributes = pca.transformedHeader();
        for (int i = 0; i < headerAttributes.numAttributes(); i++) {
            Attribute attr = headerAttributes.attribute(i);
            
            if (attr.isNominal()) {
                // Type attribute
                //System.out.println("Attribute  " +  i +  "  Ranks  " + attr.name() + " " + attr.value(0));
            } else if (attr.isNumeric()) {
                // Principal Component
                System.out.println("Component " +  i +  ":\t" + attr.name());
            }
        }
    }
    
    // Save a data set (it will add ".arff" extention automatically)
    protected void saveInstances(String outputDir, String name, Instances data) throws Exception {
        BufferedWriter writer = new BufferedWriter(new FileWriter(outputDir + "/" + name + ".arff"));
        writer.write(data.toString());
        writer.newLine();
        writer.flush();
        writer.close();
    }
    
    // Perform PCA and save results (filename should not include ".arff")
    //  - It returns -1 in case of error
    public void generateData(String inputDir, String filename, String outputDir, double varianceCovered) throws Exception {
        Instances data = readData(inputDir, filename);
        PrincipalComponents pca = applyPCA(data, varianceCovered);
        printPCA(filename + "_PCA_" + varianceCovered, pca);
        // Save principal components
        saveResults(outputDir, filename + "_PCA_" + varianceCovered, pca);
        // Transform data and save it
        saveInstances(outputDir, filename + "_PCA_" + varianceCovered, pca.transformedData(data));
    }
    
    // Default covered variance = 95%
    public void generateData(String inputDir, String filename, String outputDir) throws Exception {
        this.generateData(inputDir, filename, outputDir, 0.95);
    }


    // Main not needed because PCA class will be used from R language
	/*public static void main(String[] args) throws Exception {
	
	    PCA pca = new PCA();
        
        // Execute PCA for each file
        String[] filenames = {"00.Lineal", "01.Rotated_Lineal", "02.Small_SigmaX", "03.Large_SigmaX",  "04.Small_SigmaX_and_similar_SigmaErr"};
        
        for(String name: filenames) {
            PCA.generateData("input", name, "output");
        }

	}*/

}


