#### Functions for reading and writing ARFF files (weka format)
#### Adapted by Sergi Blanco Cuaresma from RWeka package
## License:	GPL-2
##  http://cran.r-project.org/web/packages/RWeka/index.html

########################
### Public functions
########################

###### Read ARFF file (Weka format)
read.arff = function(file) {
    ## Read the ARFF file into Weka.
    reader = .jnew("java/io/FileReader", file)
    instances = .jnew("weka/core/Instances", .jcast(reader, "java/io/Reader"))

    read_instances_from_Weka(instances)
}

###### Write ARFF file (Weka format)
write.arff = function(data, file, eol = "\n") {
    file = file(file, 'w')
    on.exit(close(file))

    if(!is.data.frame(data) && !is.matrix(data)) {
        x = data.frame(data)
    }

    instances = read_data_into_Weka(data)
    text = .jcall(instances, "Ljava/lang/String;", "toString")

    writeLines(text, file, sep = eol)
}

########################
### Support functions
########################

# Transform Weka instances to R Data Frame
read_instances_from_Weka = function(x) {
    ## Get attribute information
    out = vector("list", .jcall(x, "I", "numAttributes"))
    for (i in seq_along(out)) {
        ## In Weka missing values are coded as NaN and the cast 
        ## to double should ensure this for all attribute types.
        out[[i]] = .jcall(x, "[D", "attributeToDoubleArray", as.integer(i - 1L))
        attribute = .jcall(x, "Lweka/core/Attribute;", "attribute", as.integer(i - 1L))
        names(out)[i] = .jcall(attribute, "S", "name")
        ## See the Constant Field Values in the Weka documentation.
        switch(.jcall(attribute, "I", "type") + 1L,
           {   ## 0 numeric (nothing todo)

               is.na(out[[i]]) = is.nan(out[[i]])
           },
            
           {   ## 1 nominal (Weka value code = R level code - 1)

               idx = seq(.jcall(attribute, "I", "numValues")) - 1L
               is.na(out[[i]]) = is.nan(out[[i]])
               out[[i]] = factor(out[[i]], levels = idx)
               levels(out[[i]]) =
                   sapply(idx, function(k)
                          .jcall(attribute, "S", "value", as.integer(k)))
               ## Assume logical (see below).
               if(all(match(levels(out[[i]]), c("FALSE", "TRUE"),
                       nomatch = 0L)))
                   out[[i]] = out[[i]] == "TRUE"
           },
            
           {   ## 2 string (same as 1 but return as character)

               idx = seq(.jcall(attribute, "I", "numValues")) - 1L
               is.na(out[[i]]) = is.nan(out[[i]])
               out[[i]] = factor(out[[i]], levels = idx)
               levels(out[[i]]) =
                   sapply(idx, function(k)
                          .jcall(attribute, "S", "value", as.integer(k)))
               out[[i]] = as.character(out[[i]])
           },
            
           {   ## 3 date (the direct approach is not reliable)
               ##

               ## Format date.
               out[[i]] = .jcall("RWekaInterfaces", "[Ljava/lang/String;",
                                  "formatDate", attribute, .jarray(out[[i]]),
                                  NA_character_)
               ## Represent date in local time.
               out[[i]] = as.POSIXct(out[[i]], tz = "")
           },

           {   ## 4 relational
               stop("Type 'relational' currently not implemented.")
           },

           {   ## unknown
               stop("Type not implemented")
           }
        )
    }
    ## NOTE that Weka codes a missing class attribute as -1.
    classIndex = .jcall(x, "I", "classIndex") + 1L
    if(classIndex && classIndex != length(out)) {
        out = c(out[-classIndex], out[classIndex])
    }

    ## Prevent garbling of attribute names, etc.
    data.frame(out, check.names = FALSE, stringsAsFactors = FALSE)
}

# Transform R Data Frame to Weka instances
read_data_into_Weka = function(x, classIndex = 0L) {
    ## FastVector was deprecated in Weka >= 3-7-1. Now we have to use
    ## the List interface (see the cast of ArrayList in the Attribute
    ## constructor).

    ## See the Weka 3-5-7 source code for this insanity (e.g., string).
    ## Note that the class index, if any, must be set as an attribute.

    ## As Weka instance objects do not have case/row names, we store
    ## such information in the R container for the Weka instances.  For
    ## simplicity, we store the dimnames including the (variable) names
    ## also contained in the Weka instances.
    dx = dim(x)
    dnx = dimnames(x)

    ##################################
    ###### Build attribute information
    ##################################
    attname = names(x)
    attinfo = .jnew("weka/core/FastVector", as.integer(length(x)))
    for (i in seq_along(x)) {
        ## Make logicals into Weka nominals.
        if(is.logical(x[[i]])) {
            x[[i]] = factor(x[[i]])
        }
        attribute = 
            if(is.factor(x[[i]])) { # Type attribute
               nominalValues = .jnew("weka/core/FastVector", as.integer(nlevels(x[[i]])+1))
               for(t in levels(x[[i]])) { # Possible values of the attribute
                .jcall(nominalValues, "V", "addElement", .jcast(.jnew("java/lang/String", t)))
               }
               
               ## Add an extra ficticious category (avoid Weka GUI Bug when there is only 1 category)
               .jcall(nominalValues, "V", "addElement", .jcast(.jnew("java/lang/String", "X")))
               
               ## shift to Weka's internal coding (0 is the first possible value and not 1)
               x[[i]] = as.double(x[[i]]) - 1

                .jnew("weka/core/Attribute", attname[i], .jcast(nominalValues, "weka/core/FastVector"))
            }
#            # Code from RWeka: not addapted because it is not needed for the exercise
#            else if(is.character(x[[i]])) {
#               att = .jnew("weka/core/Attribute", attname[i],
#                            .jnull("java/util/List"))
#               x[[i]] = as.factor(x[[i]])
#               index = sapply(levels(x[[i]]), function(k)
#                               .jcall(att, "I", "addStringValue", k))
#               if(any(index < 0))
#                   stop("pushing to Type 'string' failed")
#               x[[i]] = as.double(index[as.integer(x[[i]])])

#               att
#            }
#            else if(inherits(x[[i]], "POSIXt")) {
#               att = .jnew("weka/core/Attribute", attname[i],
#                            "yyyy-MM-dd HH:mm:ss")
#               ## Normalize to local time.
#               x[[i]] = .jcall("RWekaInterfaces", "[D", "parseDate", att,
#                                .jarray(format(x[[i]], tz = "")),
#                                NA_character_)
#               att
#            }
            else if(is.numeric(x[[i]])) { # Number attribute
               .jnew("weka/core/Attribute", attname[i])
            } else {
                stop("Type not implemented")
            }
            
        .jcall(attinfo, "V", "addElement", .jcast(attribute, "java/lang/Object"))
    }
    
    ## Build instances.
    n = dim(x)[1L]                     # number of instances
    instances = .jnew("weka/core/Instances",
                       "R_data_frame",  # FIXME
                       attinfo,
                       as.integer(n))   # capacity
    
    ## Set class index.
    if(classIndex > 0L)
       .jcall(instances, "V", "setClassIndex", as.integer(classIndex - 1L))

    ##################################
    ###### Add intances (data)
    ##################################
    for(i in 1:dim(x)[1L]) {
        instance = .jnew("weka/core/Instance", as.integer(dim(x)[2]))
        for(j in 1:dim(x)[2L]) {
            .jcall(instance, "V", "setValue", as.integer(j-1), as.double(x[i,j]))
        }

        .jcall(instances, "V", "add", instance)
    }

    ## Note that using dim and dimnames attributes would result in a
    ## matrix, which seems a bad idea.
    .structure(instances, .dim = dx, .dimnames = dnx)
}

.structure = function(x, ...) {
    `attributes<-`(x, c(attributes(x), list(...)))
}
