#!/usr/bin/R
library(rJava)
library(GDD)
library(scatterplot3d)
source("WekaFileFunctions.R") ## Weka ARFF file format read and write functions

# Initialite Java Virtual Machine
.jinit(classpath="weka.jar:./")
# Create PCA object for Principal Components Analysis
pca<-.jnew("PCA")

########################
### Problem functions
########################

#### Generate a 3D sample (3 x N numbers) according to problem specifications
generateSample = function(filename, sigmaX = 1, sigmaErr = 1, sigmaZ = 1, n = 1000) {

    x = rnorm(n, mean=0, sd=sigmaX)
    y = exp(1)**x + rnorm(n, mean=0, sd=sigmaErr)
    z = rnorm(n, mean=0, sd=sigmaZ)
    type = rep("A", n)

    df = data.frame(x, y, z, type)

    write.arff(df, file = paste("output/", filename, ".arff", sep=""))
    
    df # Return data frame
}

#### Generate a 3D plot and save it to disk (output directory)
generatePlot = function(plotname, X, Y, Z) {
    GDD(file=paste("output/", plotname, sep=""), type="png", w=800, h=600)
    scatterplot3d(X, Y, Z, highlight.3d=TRUE, col.axis="blue", col.grid="lightblue", main="Sample distribution", sub=plotname, pch=20, type="p", angle=60)
    dev.off() 
}

#### Generate a histogram and save it to disk (output directory)
## Used for the last part of the exercise
generateHistogram = function(plotname, data) {
    GDD(file=paste("output/05.", plotname, sep=""), type="png", w=800, h=600)
    hist(data, probability=TRUE, breaks=50, main=paste("Histogram for", plotname), xlab="% Value change")
    rug(data)
    lines(density(data), col="red", lwd=1)
    dev.off() 
}



########################
### Tasks execution
########################

#------------------
#-- First task
#------------------
lineal = read.arff("input/00.Lineal.arff")
write.arff(lineal, file = "output/00.Lineal.arff") # Just copy to output directory
generatePlot("00.Lineal", lineal$X, lineal$Y, lineal$Z)

# Perform PCA over original linear data
varianceCovered = 0.95
.jcall(pca, "V", "generateData", "input", "00.Lineal", "output", as.double(varianceCovered))
varianceCovered = 0.70
.jcall(pca, "V", "generateData", "input", "00.Lineal", "output", as.double(varianceCovered))

# -45º rotation of the linear data around Z axis
x = lineal$X * cos(-45 * (pi/180)) - lineal$Y * sin(-45 * (pi/180))
y = lineal$X * sin(-45 * (pi/180)) - lineal$Y * cos(-45 * (pi/180))
z = lineal$Z
type = rep("A", 1000)
linealRotated = data.frame(x, y, z, type)
write.arff(linealRotated, file = "output/01.Rotated_Lineal.arff")

generatePlot("01.Rotated_Lineal", linealRotated$x, linealRotated$y, linealRotated$z)

# Perform PCA over rotated Lineal data
varianceCovered = 0.95
.jcall(pca, "V", "generateData", "output", "01.Rotated_Lineal", "output", as.double(varianceCovered))
varianceCovered = 0.70
.jcall(pca, "V", "generateData", "output", "01.Rotated_Lineal", "output", as.double(varianceCovered))

#------------------
#-- Second task
#------------------
d1 = generateSample("02.Small_SigmaX", sigmaX=0.1, sigmaErr=0.1e-5, sigmaZ=1)
d2 = generateSample("03.Large_SigmaX", sigmaX=100, sigmaErr=0.1e-5, sigmaZ=1)
d3 = generateSample("04.Small_SigmaX_and_similar_SigmaErr", sigmaX=0.1, sigmaErr=0.098, sigmaZ=1)

generatePlot("02.Small_SigmaX", d1$x, d1$y, d1$z)
generatePlot("03.Large_SigmaX", d2$x, d2$y, d2$z)
generatePlot("04.Small_SigmaX_and_similar_SigmaErr", d3$x, d3$y, d3$z)

# Perform PCA
varianceCovered = 0.95
.jcall(pca, "V", "generateData", "output", "02.Small_SigmaX", "output", as.double(varianceCovered))
.jcall(pca, "V", "generateData", "output", "03.Large_SigmaX", "output", as.double(varianceCovered))
.jcall(pca, "V", "generateData", "output", "04.Small_SigmaX_and_similar_SigmaErr", "output", as.double(varianceCovered))
varianceCovered = 0.70
.jcall(pca, "V", "generateData", "output", "02.Small_SigmaX", "output", as.double(varianceCovered))
.jcall(pca, "V", "generateData", "output", "03.Large_SigmaX", "output", as.double(varianceCovered))
.jcall(pca, "V", "generateData", "output", "04.Small_SigmaX_and_similar_SigmaErr", "output", as.double(varianceCovered))

#------------------
#-- Third task
#------------------

# Alternative variables aiming to have more linear relationships
type = rep("A", length(d1$y))
if (min(d1$y) > 0) {
    d1_alt = data.frame(x = d1$x, y = log(d1$y), z = d1$z, type)
} else {
    # Natural logarithm should have positive non-zero values to avoid problems
    # * All data will be displaced by the same quantity, it will not affect PCA
    d1_alt = data.frame(x = d1$x, y = log(d1$y - min(d1$y) + 0.1e-5), z = d1$z, type)
}
if (min(d2$y) > 0) {
    d2_alt = data.frame(x = d2$x, y = log(d2$y), z = d2$z, type)
} else {
    # Natural logarithm should have positive non-zero values to avoid problems
    # * All data will be displaced by the same quantity, it will not affect PCA
    d2_alt = data.frame(x = d2$x, y = log(d2$y - min(d2$y) + 0.1e-5), z = d2$z, type)
}
if (min(d3$y) > 0) {
    d3_alt = data.frame(x = d3$x, y = log(d3$y), z = d3$z, type)
} else {
    # Natural logarithm should have positive non-zero values to avoid problems
    # * All data will be displaced by the same quantity, it will not affect PCA
    d3_alt = data.frame(x = d3$x, y = log(d3$y - min(d3$y) + 0.1e-5), z = d3$z, type)
}

write.arff(d1_alt, file = paste("output/", "02.Small_SigmaX_ALT", ".arff", sep=""))
write.arff(d2_alt, file = paste("output/", "03.Large_SigmaX_ALT", ".arff", sep=""))
write.arff(d3_alt, file = paste("output/", "04.Small_SigmaX_and_similar_SigmaErr_ALT", ".arff", sep=""))

generatePlot("02.Small_SigmaX_ALT", d1_alt$x, d1_alt$y, d1_alt$z)
generatePlot("03.Large_SigmaX_ALT", d2_alt$x, d2_alt$y, d2_alt$z)
generatePlot("04.Small_SigmaX_and_similar_SigmaErr_ALT", d3_alt$x, d3_alt$y, d3_alt$z)

# Perform PCA
varianceCovered = 0.95
.jcall(pca, "V", "generateData", "output", "02.Small_SigmaX_ALT", "output", as.double(varianceCovered))
.jcall(pca, "V", "generateData", "output", "03.Large_SigmaX_ALT", "output", as.double(varianceCovered))
.jcall(pca, "V", "generateData", "output", "04.Small_SigmaX_and_similar_SigmaErr_ALT", "output", as.double(varianceCovered))
varianceCovered = 0.70
.jcall(pca, "V", "generateData", "output", "02.Small_SigmaX_ALT", "output", as.double(varianceCovered))
.jcall(pca, "V", "generateData", "output", "03.Large_SigmaX_ALT", "output", as.double(varianceCovered))
.jcall(pca, "V", "generateData", "output", "04.Small_SigmaX_and_similar_SigmaErr_ALT", "output", as.double(varianceCovered))

####################################
### Real data: Stock market values
####################################

# Read input data
ibex = read.table("input/IBEX.csv", header = TRUE, sep = ",", row.names = NULL)
bbva = read.table("input/BBVA.csv", header = TRUE, sep = ",", row.names = NULL)
santander = read.table("input/Santander.csv", header = TRUE, sep = ",", row.names = NULL)
banesto = read.table("input/Banesto.csv", header = TRUE, sep = ",", row.names = NULL)
bankinter = read.table("input/Bankinter.csv", header = TRUE, sep = ",", row.names = NULL)
popular = read.table("input/Popular.csv", header = TRUE, sep = ",", row.names = NULL)
telefonica = read.table("input/Telefonica.csv", header = TRUE, sep = ",", row.names = NULL)

## Just interested in 3 columns
ibex = data.frame(ibex$Date, ibex$Open, ibex$Close)
bbva = data.frame(bbva$Date, bbva$Open, bbva$Close)
santander = data.frame(santander$Date, santander$Open, santander$Close)
banesto = data.frame(banesto$Date, banesto$Open, banesto$Close)
bankinter = data.frame(bankinter$Date, bankinter$Open, bankinter$Close)
popular = data.frame(popular$Date, popular$Open, popular$Close)
telefonica = data.frame(telefonica$Date, telefonica$Open, telefonica$Close)

## Join everything excluding those days for which we do not have values for all stocks
stock = merge(bbva, santander, by.x=c("bbva.Date"), by.y=c("santander.Date"))
stock = merge(stock, banesto, by.x=c("bbva.Date"), by.y=c("banesto.Date"))
stock = merge(stock, bankinter, by.x=c("bbva.Date"), by.y=c("bankinter.Date"))
stock = merge(stock, popular, by.x=c("bbva.Date"), by.y=c("popular.Date"))
stock = merge(stock, telefonica, by.x=c("bbva.Date"), by.y=c("telefonica.Date"))
stock = merge(stock, ibex, by.x=c("bbva.Date"), by.y=c("ibex.Date"))

#---------------------------------------------
#-- Select only % change for banks
#---------------------------------------------
onlyBanksStockChange = data.frame(bbva.Change = (stock$bbva.Close - stock$bbva.Open)/stock$bbva.Open, santander.Change = (stock$santander.Close - stock$santander.Open)/stock$santander.Open, banesto.Change = (stock$banesto.Close - stock$banesto.Open)/stock$banesto.Open, bankinter.Change = (stock$bankinter.Close - stock$bankinter.Open)/stock$bankinter.Open, popular.Change = (stock$popular.Close - stock$popular.Open)/stock$popular.Open, positiveDay =  ((stock$ibex.Close - stock$ibex.Open) >= 0))

## Save data in weka format
write.arff(onlyBanksStockChange, file = "output/05.Stock.Change.only.Banks.arff")

# Perform PCA
varianceCovered = 0.95
.jcall(pca, "V", "generateData", "output", "05.Stock.Change.only.Banks", "output", as.double(varianceCovered))
varianceCovered = 0.70
.jcall(pca, "V", "generateData", "output", "05.Stock.Change.only.Banks", "output", as.double(varianceCovered))


#---------------------------------------------
#-- Select % change for banks and Telefonica
#---------------------------------------------
banksAndTelefonicaStockChange = data.frame(bbva.Change = (stock$bbva.Close - stock$bbva.Open)/stock$bbva.Open, santander.Change = (stock$santander.Close - stock$santander.Open)/stock$santander.Open, banesto.Change = (stock$banesto.Close - stock$banesto.Open)/stock$banesto.Open, bankinter.Change = (stock$bankinter.Close - stock$bankinter.Open)/stock$bankinter.Open, popular.Change = (stock$popular.Close - stock$popular.Open)/stock$popular.Open, telefonica.Change = (stock$telefonica.Close - stock$telefonica.Open)/stock$telefonica.Open, positiveDay =  ((stock$ibex.Close - stock$ibex.Open) >= 0))

## Save data in weka format
write.arff(banksAndTelefonicaStockChange, file = "output/05.Stock.Change.Banks.and.Telefonica.arff")

# Perform PCA
varianceCovered = 0.95
.jcall(pca, "V", "generateData", "output", "05.Stock.Change.Banks.and.Telefonica", "output", as.double(varianceCovered))
varianceCovered = 0.70
.jcall(pca, "V", "generateData", "output", "05.Stock.Change.Banks.and.Telefonica", "output", as.double(varianceCovered))

#-------------------------
#-- Statistical analysis
#-------------------------

# "Alias"
s = banksAndTelefonicaStockChange

file = file("output/05.Statistics.Analysis.csv", 'w')
writeLines(toString(c("Stock", "Min", "Max", "Mean", "Median", "Variance", "Std Deviation")), file)
writeLines(toString(c("BBVA", min(s$bbva.Change), max(s$bbva.Change), mean(s$bbva.Change), median(s$bbva.Change), var(s$bbva.Change), sd(s$bbva.Change))), file)
writeLines(toString(c("Santander", min(s$santander.Change), max(s$santander.Change), mean(s$santander.Change), median(s$santander.Change), var(s$santander.Change), sd(s$santander.Change))), file)
writeLines(toString(c("Banesto", min(s$banesto.Change), max(s$banesto.Change), mean(s$banesto.Change), median(s$banesto.Change), var(s$banesto.Change), sd(s$banesto.Change))), file)
writeLines(toString(c("Bankinter", min(s$bankinter.Change), max(s$bankinter.Change), mean(s$bankinter.Change), median(s$bankinter.Change), var(s$bankinter.Change), sd(s$bankinter.Change))), file)
writeLines(toString(c("Popular", min(s$popular.Change), max(s$popular.Change), mean(s$popular.Change), median(s$popular.Change), var(s$popular.Change), sd(s$popular.Change))), file)
writeLines(toString(c("Telefonica", min(s$telefonica.Change), max(s$telefonica.Change), mean(s$telefonica.Change), median(s$telefonica.Change), var(s$telefonica.Change), sd(s$telefonica.Change))), file)
close(file)

generateHistogram("BBVA", s$bbva.Change)
generateHistogram("Santander", s$santander.Change)
generateHistogram("Banesto", s$banesto.Change)
generateHistogram("Bankinter", s$bankinter.Change)
generateHistogram("Popular", s$popular.Change)
generateHistogram("Telefonica", s$telefonica.Change)

## Number of positive days
length(s$positiveDay[s$positiveDay == TRUE])
length(s$positiveDay[s$positiveDay == FALSE])

